#  Authors: Sylvain MARIE <sylvain.marie@se.com>
#            + All contributors to <https://github.com/smarie/python-genbadge>
#
#  License: 3-clause BSD, <https://github.com/smarie/python-genbadge/blob/master/LICENSE>
from __future__ import division

from io import TextIOWrapper

try:
    from typing import Union
except ImportError:
    pass

# use our own copy so as to use defusedxml and to be compliant with setuptools>=58
from .xunitparser_copy import parse
from .utils_badge import Badge


class TestStats(object):
    """
    Contains the results from parsing the Junit style test report (junit.xml).
    Some stats are available as properties, computed based on others.
    """
    def __init__(self, runned, skipped, failed, errors):
        self.runned = runned
        self.failed = failed
        self.skipped = skipped
        self.errors = errors

    @property
    def success(self):
        return self.runned - self.skipped - self.failed

    @property
    def total_with_skipped(self):
        return self.runned + self.errors

    @property
    def total_without_skipped(self):
        return self.runned - self.skipped + self.errors

    @property
    def success_percentage(self):
        if self.runned > 0:
            return self.success * 100 / self.total_without_skipped
        else:
            return 100

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, ",".join("%s=%r" % (k, v) for k, v in vars(self).items()))


def get_test_stats(junit_xml_file='reports/junit/junit.xml'  # type: Union[str, TextIOWrapper]
                   ):
    # type: (...) -> TestStats
    """
    read the junit test file and extract the success percentage
    :param junit_xml_file: the junit xml file path or file/text stream
    :return: the success percentage (an int)
    """
    if isinstance(junit_xml_file, str):
        # assume a file path
        with open(junit_xml_file) as f:
            ts, tr = parse(f)
    else:
        # assume a stream already
        ts, tr = parse(junit_xml_file)

    runned = tr.testsRun
    skipped = len(tr.skipped)
    failed = len(tr.failures)
    errors = len(tr.errors)

    return TestStats(runned=runned, skipped=skipped, failed=failed, errors=errors)


def get_color(
        test_stats  # type: TestStats
):
    """ Returns the badge color to use depending on the success percentage """

    if test_stats.success_percentage < 50:
        color = 'red'
    elif test_stats.success_percentage < 75:
        color = 'orange'
    elif test_stats.success_percentage < 90:
        color = 'green'
    else:
        color = 'brightgreen'

    return color


def get_tests_badge(
        test_stats,  # type: TestStats
        left_txt = "tests" # type: str
):
    # type: (...) -> Badge
    """Return the badge from test statistics """

    color = get_color(test_stats)

    if test_stats.success_percentage < 100:
        # full badge
        right_txt = "%s/%s" % (test_stats.success, test_stats.total_without_skipped)
    else:
        # simplified badge showing the number of tests
        right_txt = "%s" % (test_stats.total_without_skipped,)

    return Badge(left_txt=left_txt, right_txt=right_txt, color=color)



from collections import namedtuple
import logging
from pathlib import Path
import shutil
import os

from typing import Sequence, Dict, Union

import nox


nox_logger = logging.getLogger("nox")


PY27, PY35, PY36, PY37, PY38, PY39, PY310, PY311, PY312, PY313 = ("2.7", "3.5", "3.6", "3.7", "3.8", "3.9", "3.10",
                                                                "3.11", "3.12", "3.13")
DONT_INSTALL = "dont_install"


def install_reqs(
    session,
    # pre wired phases
    setup=False,
    install=False,
    tests=False,
    extras=(),
    # custom phase
    phase=None,
    phase_reqs=None,
    versions_dct=None
):
    """
    A high-level helper to install requirements from the various project files

     - pyproject.toml "[build-system] requires" (if setup=True)
     - setup.cfg "[options] setup_requires" (if setup=True)
     - setup.cfg "[options] install_requires" (if install=True)
     - setup.cfg "[options] test_requires" (if tests=True)
     - setup.cfg "[options.extras_require] <...>" (if extras=(a tuple of extras))

    Two additional mechanisms are provided in order to customize how packages are installed.

    Conda packages
    --------------
    If the session runs on a conda environment, you can add a [tool.conda] section to your pyproject.toml. This
    section should contain a `conda_packages` entry containing the list of package names that should be installed
    using conda instead of pip.

    ```
    [tool.conda]
    # Declare that the following packages should be installed with conda instead of pip
    # Note: this includes packages declared everywhere, here and in setup.cfg
    conda_packages = [
        "setuptools",
        "wheel",
        "pip"
    ]
    ```

    Version constraints
    -------------------
    In addition to the version constraints in the pyproject.toml and setup.cfg, you can specify additional temporary
    constraints with the `versions_dct` argument , for example if you know that this executes on a specific python
    version that requires special care.
    For this, simply pass a dictionary of {'pkg_name': 'pkg_constraint'} for example {"pip": ">10"}.

    """

    # Read requirements from pyproject.toml
    toml_setup_reqs, toml_use_conda_for = read_pyproject_toml()
    if setup:
        install_any(session, "pyproject.toml#build-system", toml_setup_reqs,
                    use_conda_for=toml_use_conda_for, versions_dct=versions_dct)

    # Read test requirements from setup.cfg
    setup_cfg = read_setuptools_cfg()
    if setup:
        install_any(session, "setup.cfg#setup_requires", setup_cfg.setup_requires,
                    use_conda_for=toml_use_conda_for, versions_dct=versions_dct)
    if install:
        install_any(session, "setup.cfg#install_requires", setup_cfg.install_requires,
                    use_conda_for=toml_use_conda_for, versions_dct=versions_dct)
    if tests:
        install_any(session, "setup.cfg#tests_requires", setup_cfg.tests_requires,
                    use_conda_for=toml_use_conda_for, versions_dct=versions_dct)

    for extra in extras:
        install_any(session, "setup.cfg#extras_require#%s" % extra, setup_cfg.extras_require[extra],
                    use_conda_for=toml_use_conda_for, versions_dct=versions_dct)

    if phase is not None:
        install_any(session, phase, phase_reqs, use_conda_for=toml_use_conda_for, versions_dct=versions_dct)


def install_any(session,
                phase_name: str,
                pkgs: Sequence[str],
                use_conda_for: Sequence[str] = (),
                versions_dct: Dict[str, str] = None,
                ):
    """Install the `pkgs` provided with `session.install(*pkgs)`, except for those present in `use_conda_for`"""

    # use the provided versions dictionary to update the versions
    if versions_dct is None:
        versions_dct = dict()
    pkgs = [pkg + versions_dct.get(pkg, "") for pkg in pkgs if versions_dct.get(pkg, "") != DONT_INSTALL]

    nox_logger.debug("\nAbout to install *%s* requirements: %s.\n "
                     "Conda pkgs are %s" % (phase_name, pkgs, use_conda_for))

    # install on conda... if the session uses conda backend
    if not isinstance(session.virtualenv, nox.virtualenv.CondaEnv):
        conda_pkgs = []
    else:
        conda_pkgs = [pkg_req for pkg_req in pkgs if any(get_req_pkg_name(pkg_req) == c for c in use_conda_for)]
        if len(conda_pkgs) > 0:
            nox_logger.info("[%s] Installing requirements with conda: %s" % (phase_name, conda_pkgs))
            session.conda_install(*conda_pkgs)

    pip_pkgs = [pkg_req for pkg_req in pkgs if pkg_req not in conda_pkgs]
    # safety: make sure that nothing went modified or forgotten
    assert set(conda_pkgs).union(set(pip_pkgs)) == set(pkgs)
    if len(pip_pkgs) > 0:
        nox_logger.info("[%s] Installing requirements with pip: %s" % (phase_name, pip_pkgs))
        session.install(*pip_pkgs)


# ------------- requirements related


def read_pyproject_toml() -> Union[list, list]:
    """
    Reads the `pyproject.toml` and returns

     - a list of setup requirements from [build-system] requires
     - sub-list of these requirements that should be installed with conda, from [tool.my_conda] conda_packages
    """
    if os.path.exists("pyproject.toml"):
        import toml
        nox_logger.debug("\nA `pyproject.toml` file exists. Loading it.")
        pyproject = toml.load("pyproject.toml")
        requires = pyproject['build-system']['requires']
        try:
            conda_pkgs = pyproject['tool']['conda']['conda_packages']
        except KeyError:
            conda_pkgs = []
        return requires, conda_pkgs
    else:
        raise FileNotFoundError("No `pyproject.toml` file exists. No dependency will be installed ...")


SetupCfg = namedtuple('SetupCfg', ('setup_requires', 'install_requires', 'tests_requires', 'extras_require'))


def read_setuptools_cfg():
    """
    Reads the `setup.cfg` file and extracts the various requirements lists
    """
    # see https://stackoverflow.com/a/30679041/7262247
    from setuptools import Distribution
    dist = Distribution()
    dist.parse_config_files()
    return SetupCfg(setup_requires=dist.setup_requires,
                    install_requires=dist.install_requires,
                    tests_requires=dist.tests_require,
                    extras_require=dist.extras_require)


def get_req_pkg_name(r):
    """Return the package name part of a python package requirement.

    For example
    "funcsigs;python<'3.5'" will return "funcsigs"
    "pytest>=3" will return "pytest"
    """
    return r.replace('<', '=').replace('>', '=').replace(';', '=').split("=")[0]


# ----------- other goodies


def rm_file(folder: Union[str, Path]):
    """Since on windows Path.unlink throws permission error sometimes, os.remove is preferred."""
    if isinstance(folder, str):
        folder = Path(folder)

    if folder.exists():
        os.remove(str(folder))
        # Folders.site.unlink()  --> possible PermissionError


def rm_folder(folder: Union[str, Path]):
    """Since on windows Path.unlink throws permission error sometimes, shutil is preferred."""
    if isinstance(folder, str):
        folder = Path(folder)

    if folder.exists():
        shutil.rmtree(str(folder))
        # Folders.site.unlink()  --> possible PermissionError



#  Authors: Sylvain MARIE <sylvain.marie@se.com>
#            + All contributors to <https://github.com/smarie/python-genbadge>
#
#  License: 3-clause BSD, <https://github.com/smarie/python-genbadge/blob/master/LICENSE>
from __future__ import division

from .utils_badge import Badge

try:
    # security patch: see https://docs.python.org/3/library/xml.etree.elementtree.html
    import defusedxml.ElementTree as defused_etree
except ImportError as e:
    ee = e  # save it
    class FakeDefusedXmlImport(object):  # noqa
        def __getattribute__(self, item):
            raise ImportError("Could not import `defusedxml.ElementTree`, please install `defusedxml`. "
                              "Note that all dependencies for the coverage command can be installed with "
                              "`pip install genbadge[coverage]`. Caught: %r" % ee)
    defused_etree = FakeDefusedXmlImport()


class CoverageStats(object):
    """
    Contains the results from parsing the coverage.xml.
    """
    def __init__(self,
                 branches_covered=None, branches_valid=None, branch_option=None,
                 complexity=None, lines_covered=None, lines_valid=None,
                 ):
        self.complexity = complexity

        self.branches_covered = branches_covered
        self.branches_valid = branches_valid
        self.branch_option = branch_option

        self.lines_covered = lines_covered
        self.lines_valid = lines_valid

    @property
    def branch_rate(self):
        """
        Note: in --no-branch situations, the number of branches is 0.
        In that case, the branch rate is 0 in the coverage.xml.
        But in --branch situations without actual branches,
        the number of branches is also 0 but the branch rate is 1.
        We mimic both behaviours in this field to be consistent.
        """
        if self.branches_valid > 0:
            return self.branches_covered / self.branches_valid
        elif self.branch_option:
            return 1
        else:
            return 0

    @property
    def line_rate(self):
        """See branch rate for the special case of division by zero"""
        if self.lines_valid > 0:
            return self.lines_covered / self.lines_valid
        else:
            return 0

    @property
    def branch_coverage(self):
        return self.branch_rate * 100

    @property
    def line_coverage(self):
        return self.line_rate * 100

    @property
    def total_rate(self):
        """
        See XmlReport class in https://github.com/nedbat/coveragepy/blob/master/coverage/xmlreport.py
        for the formula.

        See branch rate for the special case of division by zero.
        """
        denom = self.lines_valid + self.branches_valid
        if denom > 0:
            return (self.lines_covered + self.branches_covered) / denom
        else:
            return 0

    @property
    def total_coverage(self):
        return self.total_rate * 100


def get_coverage_stats(coverage_xml_file):
    # type: (...) -> CoverageStats
    """
    Reads a coverage.xml file

    <coverage branch-rate="0.6" branches-covered="24" branches-valid="40" complexity="0" line-rate="0.8586"
              lines-covered="170" lines-valid="198" timestamp="1620747625339" version="5.5">
    </coverage>
    """
    if isinstance(coverage_xml_file, str):
        # assume a file path
        with open(coverage_xml_file) as f:
            cov_stats = parse_cov(f)
    else:
        # assume a stream already
        cov_stats = parse_cov(coverage_xml_file)

    return cov_stats


def get_color(
        cov_stats  # type: CoverageStats
):
    """ Returns the badge color to use depending on the coverage rate """

    if cov_stats.total_coverage < 50:
        color = 'red'
    elif cov_stats.total_coverage < 75:
        color = 'orange'
    elif cov_stats.total_coverage < 90:
        color = 'green'
    else:
        color = 'brightgreen'

    return color


def get_coverage_badge(
    cov_stats,  # type: CoverageStats
    left_txt= "coverage" # type: str
):
    # type: (...) -> Badge
    """Return the badge from coverage results """

    color = get_color(cov_stats)

    right_txt = "%.2f%%" % (cov_stats.total_coverage,)

    return Badge(left_txt=left_txt, right_txt=right_txt, color=color)


def parse_cov(source):
    """Parses the coverage.xml contents from source"""
    return CovParser().parse(source)


class CovParser(object):
    """Parser class - inspired by the code in `xunitparser`"""

    def parse(self, source):
        xml = defused_etree.parse(source)
        root = xml.getroot()
        return self.parse_root(root)

    def parse_root(self, root):
        cov = CoverageStats()
        assert root.tag == 'coverage'

        cov.complexity = float(root.attrib.get('complexity'))

        cov.branches_covered = int(root.attrib.get('branches-covered'))
        cov.branches_valid = int(root.attrib.get('branches-valid'))

        cov.lines_covered = int(root.attrib.get('lines-covered'))
        cov.lines_valid = int(root.attrib.get('lines-valid'))

        # recompute the rates for more precision, but make sure that's correct
        branch_rate = float(root.attrib.get('branch-rate'))
        line_rate = float(root.attrib.get('line-rate'))

        # detect whether the --branch option were set or not
        # so CoverageStats knows how to distinguish between them
        cov.branch_option = cov.branches_valid > 0 or branch_rate == 1.0

        if not is_close(cov.branch_rate, branch_rate):
            raise ValueError("Computed branch rate (%s) is different from the one in the file (%s)"
                             % (cov.branch_rate, branch_rate))
        if not is_close(cov.line_rate, line_rate):
            raise ValueError("Computed line rate (%s) is different from the one in the file (%s)"
                             % (cov.line_rate, line_rate))

        # for el in root:
        #     if el.tag == 'sources':
        #         self.parse_sources(el, ts)
        #     if el.tag == 'packages':
        #         self.parse_packages(el, ts)

        return cov


def is_close(a, b):
    """Return True if there is at most a difference of 1 at the 2d decimal"""
    return abs(a - b) <= 0.01



#  Authors: Sylvain MARIE <sylvain.marie@se.com>
#            + All contributors to <https://github.com/smarie/python-genbadge>
#
#  License: 3-clause BSD, <https://github.com/smarie/python-genbadge/blob/master/LICENSE>
from __future__ import division

from warnings import warn
import re

from .utils_badge import Badge


try:
    # flake8-html is an optional dependency, do not fail too soon if it cant be loaded
    import flake8_html
except ImportError as e:
    ee = e  # save it
    class FakeFlake8HtmlImport(object):  # noqa
        def __getattribute__(self, item):
            raise ImportError("Could not import `flake8_html` module, please install it. "
                              "Note that all dependencies for the flake8 command can be installed with "
                              "`pip install genbadge[flake8]`. Caught: %r" % ee)
    flake8_html = FakeFlake8HtmlImport()


class Flake8Stats(object):
    """
    Contains the results from parsing the flake8 report.
    The severity levels are defined by flake8-html
    """
    def __init__(self,
                 nb_critical=0, nb_warning=0, nb_info=0
                 ):
        # severities 1, 2, 3
        self.nb_critical = nb_critical
        self.nb_warning = nb_warning
        self.nb_info = nb_info

    def add(self,
            nb,   # type: int
            code  # type: str
            ):
        """
        Add `nb` errors with the same code to the statistics.
        """
        severity = flake8_html.plugin.find_severity(code)
        if severity == 1:
            self.nb_critical += nb
        elif severity == 2:
            self.nb_warning += nb
        elif severity == 3:
            self.nb_info += nb
        else:
            raise ValueError("Unknown severity: %r for code %r" % (severity, code))

    @property
    def nb_total(self):
        return self.nb_critical + self.nb_warning + self.nb_info


def get_color(
        flake8_stats  # type: Flake8Stats
):
    """ Returns the badge color to use depending on the flake8 results """

    if flake8_stats.nb_critical > 0:
        color = 'red'
    elif flake8_stats.nb_warning > 0:
        color = 'orange'
    elif flake8_stats.nb_info > 0:
        color = 'green'
    else:
        color = 'brightgreen'

    return color


def get_flake8_badge(
    flake8_stats,  # type: Flake8Stats
    left_txt = "flake8" # type: str
):
    # type: (...) -> Badge
    """Return the badge from coverage results """

    color = get_color(flake8_stats)

    right_txt = "%s C, %s W, %s I" % (flake8_stats.nb_critical, flake8_stats.nb_warning, flake8_stats.nb_info)

    return Badge(left_txt=left_txt, right_txt=right_txt, color=color)


def get_flake8_stats(flake8_stats_file):
    # type: (...) -> Flake8Stats
    """
    Reads an index.html file obtained from flake8-html.
    """
    if isinstance(flake8_stats_file, str):
        # assume a file path
        with open(flake8_stats_file) as f:
            flake8_stats_txt = f.read()
    else:
        # assume a stream already
        flake8_stats_txt = flake8_stats_file.read()

    return parse_flake8_stats(flake8_stats_txt)


RE_TO_MATCH = re.compile(r"([0-9]+)\s+([A-Z0-9]+)\s.*")


def parse_flake8_stats(stats_txt  # type: str
                       ):
    # type: (...) -> Flake8Stats

    stats = Flake8Stats()
    for line in stats_txt.splitlines():
        match = RE_TO_MATCH.match(line)
        if not match:
            warn("Line in Flake8 statistics report does not match template and will be ignored: %r" % line)
        else:
            nb, code = match.groups()
            stats.add(int(nb), code)

    return stats


# def parse_flake8_html(html  # type: str
#                       ):
#     #
#     """Reads the flake8 html report"""
#     soup = bs4.BeautifulSoup(html, "html.parser")
#
#     # check title
#     title = soup.head.title.get_text()
#     assert title == 'flake8 violations', "Invalid flake8 html report found, unexpected title: %s" % title
#
#     # get page div
#     pagediv = soup.body.find("div", {"id": "page"})
#     assert pagediv.h1.get_text() == 'flake8 violations'
#
#     results_dct = dict()
#     ul_violations = pagediv.ul
#     for li in ul_violations.find_all('li'):
#         # synthesis
#         typ_str, severity_str = li.a.span['class']
#         assert typ_str == 'count'
#         assert severity_str.startswith('sev-')
#         count = int(li.a.span.get_text().strip())
#         worst_severity_nb = int(severity_str[4:])
#
#         count2, file_name = list(li.stripped_strings)
#         assert int(count2) == count
#
#         # we need to access the details because the count is not correct
#         li_href = li.a['href']
#         child_results_dct = parse_child_html(path, li_href)
#         for c_severity_nb, c_count in child_results_dct.items():
#             try:
#                 results_dct[c_severity_nb] += c_count
#             except KeyError:
#                 results_dct[c_severity_nb] = c_count
#
#         assert worst_severity_nb == min(child_results_dct.keys())
#         assert count == sum(child_results_dct.values())
#
#     return results_dct

#
# def parse_child_html(path,   # type: str
#                      suffix  # type: str
#                      ):
#     with open(path + suffix) as f:
#         html_child = f.read()
#     soup_child = bs4.BeautifulSoup(html_child, "html.parser")
#
#     # check title
#     title = soup_child.head.title.get_text()
#     assert title.startswith('flake8 violations'), "Invalid flake8 html report found, unexpected title: %s" % title
#
#     # get page div
#     pagediv = soup_child.body.find("div", {"id": "page"})
#     # assert pagediv.h1.get_text() == 'flake8 violations'
#
#     results_dct = dict()
#     ul_violations = pagediv.ul
#     for li in ul_violations.find_all('li', recursive=False):
#         code = li.a['data-code']  # F401, etc.
#         typ_str, severity_str = li.a.span['class']
#         assert typ_str == 'count'
#         assert severity_str.startswith('sev-')
#         count = int(li.a.span.get_text().strip())
#         severity_nb = int(severity_str[4:])
#
#         try:
#             results_dct[severity_nb] += count
#         except KeyError:
#             results_dct[severity_nb] = count
#
#     return results_dct



""" A copy of xunitparser, to temporary fix the issue in https://github.com/smarie/python-genbadge/issues/18 """
import math
import unittest
from datetime import timedelta

try:
    # security patch: see https://docs.python.org/3/library/xml.etree.elementtree.html
    from defusedxml import ElementTree
except ImportError as e:
    ee = e  # save it
    class FakeDefusedXmlImport(object):  # noqa
        def __getattribute__(self, item):
            raise ImportError("Could not import `defusedxml.ElementTree`, please install `defusedxml`. "
                              "Note that all dependencies for the tests command can be installed with "
                              "`pip install genbadge[tests]`. Caught: %r" % ee)
    ElementTree = FakeDefusedXmlImport()


def to_timedelta(val):
    if val is None:
        return None

    secs = float(val)
    if math.isnan(secs):
        return None

    return timedelta(seconds=secs)


class TestResult(unittest.TestResult):
    def _exc_info_to_string(self, err, test):
        err = (e for e in err if e)
        return ': '.join(err)


class TestCase(unittest.TestCase):
    TR_CLASS = TestResult
    stdout = None
    stderr = None

    def __init__(self, classname, methodname):
        super(TestCase, self).__init__()
        self.classname = classname
        self.methodname = methodname

    def __str__(self):
        return "%s (%s)" % (self.methodname, self.classname)

    def __repr__(self):
        return "<%s testMethod=%s>" % \
               (self.classname, self.methodname)

    def __hash__(self):
        return hash((type(self), self.classname, self.methodname))

    def id(self):
        return "%s.%s" % (self.classname, self.methodname)

    def seed(self, result, typename=None, message=None, trace=None):
        """ Provide the expected result """
        self.result, self.typename, self.message, self.trace = result, typename, message, trace

    def run(self, tr=None):
        """ Fake run() that produces the seeded result """
        tr = tr or self.TR_CLASS()

        tr.startTest(self)
        if self.result == 'success':
            tr.addSuccess(self)
        elif self.result == 'skipped':
            tr.addSkip(self, '%s: %s' % (self.typename, self._textMessage()))
        elif self.result == 'error':
            tr.addError(self, (self.typename, self._textMessage()))
        elif self.result == 'failure':
            tr.addFailure(self, (self.typename, self._textMessage()))
        tr.stopTest(self)

        return tr

    def _textMessage(self):
        msg = (e for e in (self.message, self.trace) if e)
        return '\n\n'.join(msg) or None

    @property
    def alltext(self):
        err = (e for e in (self.typename, self.message) if e)
        err = ': '.join(err)
        txt = (e for e in (err, self.trace) if e)
        return '\n\n'.join(txt) or None

    def setUp(self):
        """ Dummy method so __init__ does not fail """
        pass

    def tearDown(self):
        """ Dummy method so __init__ does not fail """
        pass

    def runTest(self):
        """ Dummy method so __init__ does not fail """
        self.run()

    @property
    def basename(self):
        return self.classname.rpartition('.')[2]

    @property
    def success(self):
        return self.result == 'success'

    @property
    def skipped(self):
        return self.result == 'skipped'

    @property
    def failed(self):
        return self.result == 'failure'

    @property
    def errored(self):
        return self.result == 'error'

    @property
    def good(self):
        return self.skipped or self.success

    @property
    def bad(self):
        return not self.good

    @property
    def stdall(self):
        """ All system output """
        return '\n'.join([out for out in (self.stdout, self.stderr) if out])


class TestSuite(unittest.TestSuite):
    def __init__(self, *args, **kwargs):
        super(TestSuite, self).__init__(*args, **kwargs)
        self.properties = {}
        self.stdout = None
        self.stderr = None


class Parser(object):
    TC_CLASS = TestCase
    TS_CLASS = TestSuite
    TR_CLASS = TestResult

    def parse(self, source):
        xml = ElementTree.parse(source)
        root = xml.getroot()
        return self.parse_root(root)

    def parse_root(self, root):
        ts = self.TS_CLASS()
        if root.tag == 'testsuites':
            for subroot in root:
                self.parse_testsuite(subroot, ts)
        else:
            self.parse_testsuite(root, ts)

        tr = ts.run(self.TR_CLASS())

        tr.time = to_timedelta(root.attrib.get('time'))

        # check totals if they are in the root XML element
        if 'errors' in root.attrib:
            assert len(tr.errors) == int(root.attrib['errors'])
        if 'failures' in root.attrib:
            assert len(tr.failures) == int(root.attrib['failures'])
        if 'skip' in root.attrib:
            assert len(tr.skipped) == int(root.attrib['skip'])
        if 'tests' in root.attrib:
            assert len(list(ts)) == int(root.attrib['tests'])

        return (ts, tr)

    def parse_testsuite(self, root, ts):
        assert root.tag == 'testsuite'
        ts.name = root.attrib.get('name')
        ts.package = root.attrib.get('package')
        for el in root:
            if el.tag == 'testcase':
                self.parse_testcase(el, ts)
            if el.tag == 'properties':
                self.parse_properties(el, ts)
            if el.tag == 'system-out' and el.text:
                ts.stdout = el.text.strip()
            if el.tag == 'system-err' and el.text:
                ts.stderr = el.text.strip()

    def parse_testcase(self, el, ts):
        tc_classname = el.attrib.get('classname') or ts.name
        tc = self.TC_CLASS(tc_classname, el.attrib['name'])
        tc.seed('success', trace=el.text or None)
        tc.time = to_timedelta(el.attrib.get('time'))
        message = None
        text = None
        for e in el:
            # error takes over failure in JUnit 4
            if e.tag in ('failure', 'error', 'skipped'):
                tc = self.TC_CLASS(tc_classname, el.attrib['name'])
                result = e.tag
                typename = e.attrib.get('type')

                # reuse old if empty
                message = e.attrib.get('message') or message
                text = e.text or text

                tc.seed(result, typename, message, text)
                tc.time = to_timedelta(el.attrib.get('time'))
            if e.tag == 'system-out' and e.text:
                tc.stdout = e.text.strip()
            if e.tag == 'system-err' and e.text:
                tc.stderr = e.text.strip()

        # add either the original "success" tc or a tc created by elements
        ts.addTest(tc)

    def parse_properties(self, el, ts):
        for e in el:
            if e.tag == 'property':
                assert e.attrib['name'] not in ts.properties
                ts.properties[e.attrib['name']] = e.attrib['value']


def parse(source):
    return Parser().parse(source)



#  Authors: Sylvain MARIE <sylvain.marie@se.com>
#            + All contributors to <https://github.com/smarie/python-genbadge>
#
#  License: 3-clause BSD, <https://github.com/smarie/python-genbadge/blob/master/LICENSE>
try:
    from pathlib import Path
except ImportError:  # pragma: no cover
    from pathlib2 import Path  # python 2

import click


from .utils_junit import get_test_stats, get_tests_badge
from .utils_coverage import get_coverage_badge, get_coverage_stats
from .utils_flake8 import get_flake8_stats, get_flake8_badge

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


INFILE_HELP_TMP = "An alternate %s file to read. '-' is supported and means <stdin>."
OUTFILE_BADGE_HELP = ("An alternate SVG badge file to write to. '-' is supported and means <stdout>. Note that in this "
                      "case no other message will be printed to <stdout>. In particular the verbose flag will have no "
                      "effect.")
NAME_HELP = ("An alternate SVG badge text name to display on the left-hand side of the badge.")
WITH_NAME_HELP = ("Indicates if a badge should be generated with or without the left-hand side of the badge.")
SHIELDS_HELP = ("Indicates if badges should be generated using the shields.io HTTP API (default) or the local SVG file "
                "template included.")
VERBOSE_HELP = ("Use this flag to print details to stdout during the badge generation process. Note that this flag has "
                "no effect when '-' is used as output, since the badge is written to <stdout>. It also has no effect "
                "when the silent flag `-s` is used.")
SILENT_HELP = ("When this flag is active nothing will be written to stdout. Note that this flag has no effect when '-' "
               "is used as the output file.")


@click.group()
def genbadge():
    """
    Commandline utility to generate badges.
    To get help on each command use:

        genbadge <cmd> --help

    """
    pass


@genbadge.command(name="tests",
                  short_help="Generate a badge for the test results (e.g. from a junit.xml).")
@click.option('-i', '--input-file', type=click.File('rt'), help=INFILE_HELP_TMP % "test results XML")
@click.option('-o', '--output-file', type=click.File('wt'), help=OUTFILE_BADGE_HELP)
@click.option('-n', '--name', type=str, default="tests", help=NAME_HELP)
@click.option('-t', '--threshold', type=float,
              help="An optional success percentage threshold to use. The command will fail with exit code 1 if the"
                   "actual success percentage is strictly less than the provided value.")
@click.option('--withname/--noname', type=bool, default=True, help=WITH_NAME_HELP)
@click.option('-w/-l', '--webshields/--local', type=bool, default=True, help=SHIELDS_HELP)
@click.option('-v', '--verbose', type=bool, default=False, is_flag=True, help=VERBOSE_HELP)
@click.option('-s', '--silent', type=bool, default=False, is_flag=True, help=SILENT_HELP)
def gen_tests_badge(
        input_file=None,
        output_file=None,
        name=None,
        threshold=None,
        withname=None,
        webshields=None,
        verbose=None,
        silent=None
):
    """
    This command generates a badge for the test results, from an XML file in the
    junit format. Such a file can be for example generated from python pytest
    using the --junitxml flag, or from java junit.

    By default the input file is the relative `./reports/junit/junit.xml` and
    the output file is `./tests-badge.svg`. You can change these settings with
    the `-i/--input_file` and `-o/--output-file` options.

    By default the badge will have the name "tests" as the left-hand side text.
    You can change these settings with the `-n/--name` option. The left-hand side
    text can be left blank with `-n ""` or have the left-hand side of the badge
    completely removed by passing `--noname`.

    You can use the verbose flag `-v/--verbose` to display information on the
    input file contents, for verification.

    The resulting badge will by default look like this: [tests | 6/12]
    where 6 is the number of tests that have run successfully, and 12 is the
    total number of tests minus the number of skipped tests. When all tests
    pass with success, the badge simply shows the number of tests [tests | 12].
    You can change the appearance of the badge with the --format option (not
    implemented, todo).

    The success percentage is defined as 6/12 = 50.0%. You can use the
    `-t/--threshold` flag to setup a minimum success percentage required. If the
    success percentage is below the threshold, an error will be raised and the
    badge will not be generated.
    """
    # Process i/o files
    input_file, input_file_path = _process_infile(input_file, "reports/junit/junit.xml")
    output_file, output_file_path, is_stdout = _process_outfile(output_file, "tests-badge.svg")

    # First retrieve the success percentage from the junit xml
    try:
        test_stats = get_test_stats(junit_xml_file=input_file)
    except FileNotFoundError:
        raise click.exceptions.FileError(input_file, hint="File not found")

    if not silent and verbose and not is_stdout:
        click.echo("""Test statistics parsed successfully from %r
 - Nb tests: Total (%s) = Success (%s) + Skipped (%s) + Failed (%s) + Errors (%s)
 - Success percentage: %.2f%% (%s / %s) (Skipped tests are excluded)
""" % (input_file_path, test_stats.total_with_skipped, test_stats.success, test_stats.skipped, test_stats.failed,
       test_stats.errors, test_stats.success_percentage, test_stats.success, test_stats.total_without_skipped))

    # sanity check
    if test_stats.total_with_skipped != test_stats.success + test_stats.skipped + test_stats.failed + test_stats.errors:
        raise click.exceptions.ClickException(
            "Inconsistent junit results: the sum of all kind of tests is not equal to the total. Please report this "
            "issue if you think your file is correct. Details: %r" % test_stats
        )

    # Validate against the threshold
    if threshold is not None and test_stats.success_percentage < threshold:
        raise click.exceptions.ClickException(
            "Success percentage %s%% is strictly lower than required threshold %s%%"
            % (float(test_stats.success_percentage), threshold)
        )
        
    # Set badge name
    clear_left_txt = False
    if not withname:
        name = "" # removes left side of badge
    elif not name.strip():
        name = "###" # blank text to replace
        clear_left_txt = True # keep left side of badge but remove text

    # Generate the badge
    badge = get_tests_badge(test_stats, name)
    badge.write_to(
        output_file if is_stdout else output_file_path, 
        use_shields=webshields, 
        clear_left_txt=clear_left_txt
    )

    if not silent and not is_stdout:
        click.echo("SUCCESS - Tests badge created: %r" % str(output_file_path))


@genbadge.command(name="coverage",
                  short_help="Generate a badge for the coverage results (e.g. from a coverage.xml).")
@click.option('-i', '--input-file', type=click.File('rt'), help=INFILE_HELP_TMP % "coverage results XML")
@click.option('-o', '--output-file', type=click.File('wt'), help=OUTFILE_BADGE_HELP)
@click.option('-n', '--name', type=str, default="coverage", help=NAME_HELP)
@click.option('--withname/--noname', type=bool, default=True, help=WITH_NAME_HELP)
@click.option('-w/-l', '--webshields/--local', type=bool, default=True, help=SHIELDS_HELP)
@click.option('-v', '--verbose', type=bool, default=False, is_flag=True, help=VERBOSE_HELP)
@click.option('-s', '--silent', type=bool, default=False, is_flag=True, help=SILENT_HELP)
def gen_coverage_badge(
        input_file=None,
        output_file=None,
        name=None,
        withname=None,
        webshields=None,
        verbose=None,
        silent=None
):
    """
    This command generates a badge for the coverage results, from an XML file in
    the 'coverage' format. Such a file can be for example generated using the
    python `coverage` tool, or java `cobertura`.

    By default the input file is the relative `./reports/coverage/coverage.xml`
    and the output file is `./coverage-badge.svg`. You can change these settings
    with the `-i/--input_file` and `-o/--output-file` options.
    
    By default the badge will have the name "coverage" as the left-hand side text.
    You can change these settings with the `-n/--name` option. The left-hand side
    text can be left blank with `-n ""` or have the left-hand side of the badge
    completely removed by passing `--noname`.

    You can use the verbose flag `-v/--verbose` to display information on the
    input file contents, for verification.

    The resulting badge will by default look like this: [coverage | 98.1%] where
    98.1 is the total coverage, obtained from the branch and line coverages
    using the formula

        (nb_lines_covered + nb_branches_covered) / (nb_lines / nb_branches)

    and multiplying this by 100.
    """
    # Process i/o files
    input_file, input_file_path = _process_infile(input_file, "reports/coverage/coverage.xml")
    output_file, output_file_path, is_stdout = _process_outfile(output_file, "coverage-badge.svg")

    # First retrieve the coverage info from the coverage xml
    try:
        cov_stats = get_coverage_stats(coverage_xml_file=input_file)
    except FileNotFoundError:
        raise click.exceptions.FileError(input_file, hint="File not found")

    if not silent and verbose and not is_stdout:
        click.echo("""Coverage results parsed successfully from %(ifp)r
 - Branch coverage: %(bcp).2f%% (%(bc)s/%(bv)s)
 - Line coverage: %(lcp).2f%% (%(lc)s/%(lv)s)
 - Total coverage: %(tcp).2f%% ((%(bc)s+%(lc)s)/(%(bv)s+%(lv)s))
""" % dict(ifp=input_file_path, tcp=cov_stats.total_coverage,
           bcp=cov_stats.branch_coverage, bc=cov_stats.branches_covered, bv=cov_stats.branches_valid,
           lcp=cov_stats.line_coverage, lc=cov_stats.lines_covered, lv=cov_stats.lines_valid))

    # Set badge name
    clear_left_txt = False
    if not withname:
        name = "" # removes left side of badge
    elif not name.strip():
        name = "###" # blank text to replace
        clear_left_txt = True # keep left side of badge but remove text
    
    # Generate the badge
    badge = get_coverage_badge(cov_stats, name)    
    badge.write_to(
        output_file if is_stdout else output_file_path, 
        use_shields=webshields, 
        clear_left_txt=clear_left_txt
    )

    if not silent and not is_stdout:
        click.echo("SUCCESS - Coverage badge created: %r" % str(output_file_path))


@genbadge.command(name="flake8",
                  short_help="Generate a badge for the flake8 results (e.g. from a flake8stats.txt file).")
@click.option('-i', '--input-file', type=click.File('rt'), help=INFILE_HELP_TMP % "flake8 results TXT")
@click.option('-o', '--output-file', type=click.File('wt'), help=OUTFILE_BADGE_HELP)
@click.option('-n', '--name', type=str, default="flake8", help=NAME_HELP)
@click.option('--withname/--noname', type=bool, default=True, help=WITH_NAME_HELP)
@click.option('-w/-l', '--webshields/--local', type=bool, default=True, help=SHIELDS_HELP)
@click.option('-v', '--verbose', type=bool, default=False, is_flag=True, help=VERBOSE_HELP)
@click.option('-s', '--silent', type=bool, default=False, is_flag=True, help=SILENT_HELP)
def gen_flake8_badge(
        input_file=None,
        output_file=None,
        name=None,
        withname=None,
        webshields=None,
        verbose=None,
        silent=None
):
    """
    This command generates a badge for the flake8 results, from a flake8stats.txt
    file. Such a file can be generated from python `flake8` using the
    --statistics flag.

    By default the input file is the relative `./reports/flake8/flake8stats.txt`
    and the output file is `./flake8-badge.svg`. You can change these settings
    with the `-i/--input_file` and `-o/--output-file` options.

    By default the badge will have the name "flake8" as the left-hand side text.
    You can change these settings with the `-n/--name` option. The left-hand side
    text can be left blank with `-n ""` or have the left-hand side of the badge
    completely removed by passing `--noname`.

    You can use the verbose flag `-v/--verbose` to display information on the
    input file contents, for verification.

    The resulting badge will by default look like this: [flake8 | 6 C, 0 W, 5 I]
    where 6, 0, 5 denote the number of critical issues, warnings, and
    information messages respectively. These severity levels are determined by
    the flake8-html plugin so as to match the colors in the HTML report. You can
    change the appearance of the badge with the --format option (not
    implemented, todo).
    """
    # Process i/o files
    input_file, input_file_path = _process_infile(input_file, "reports/flake8/flake8stats.txt")
    output_file, output_file_path, is_stdout = _process_outfile(output_file, "flake8-badge.svg")

    # First retrieve the success percentage from the junit xml
    try:
        flake8_stats = get_flake8_stats(flake8_stats_file=input_file)
    except FileNotFoundError:
        raise click.exceptions.FileError(input_file, hint="File not found")

    if not silent and verbose and not is_stdout:
        click.echo("""Flake8 statistics parsed successfully from %r
 - Total (%s) = Critical (%s) + Warning (%s) + Info (%s)
""" % (input_file_path, flake8_stats.nb_total, flake8_stats.nb_critical, flake8_stats.nb_warning, flake8_stats.nb_info))

    # Set badge name
    clear_left_txt = False
    if not withname:
        name = "" # removes left side of badge
    elif not name.strip():
        name = "###" # blank text to replace
        clear_left_txt = True # keep left side of badge but remove text

    # Generate the badge
    badge = get_flake8_badge(flake8_stats, name)
    badge.write_to(
        output_file if is_stdout else output_file_path, 
        use_shields=webshields, 
        clear_left_txt=clear_left_txt
    )

    if not silent and not is_stdout:
        click.echo("SUCCESS - Flake8 badge created: %r" % str(output_file_path))


def _process_infile(input_file, default_in_file):
    """Common in file processor"""

    if input_file is None:
        input_file = default_in_file

    if isinstance(input_file, str):
        input_file_path = Path(input_file).absolute().as_posix()
    else:
        input_file_path = getattr(input_file, "name", "<stdin>")

    return input_file, input_file_path


def _process_outfile(output_file, default_out_file):
    """Common out file processor"""

    is_stdout = False
    if output_file is None:
        output_file_path = Path(default_out_file).absolute()
    elif isinstance(output_file, str):
        output_file_path = Path(output_file).absolute()
        # special case of a directory
        if output_file_path.is_dir():
            output_file_path = output_file_path / default_out_file
    else:
        output_file_path = getattr(output_file, "name", "<stdout>")
        if output_file_path == "<stdout>":
            is_stdout = True
        else:
            output_file_path = Path(output_file_path).absolute()

    if not is_stdout:
        output_file_path = output_file_path.as_posix()

    return output_file, output_file_path, is_stdout


if __name__ == '__main__':
    genbadge()



import platform
import sys
from shutil import copy

import click
from distutils.version import LooseVersion

import pytest

try:
    from pathlib import Path
except ImportError:  # pragma: no cover
    from pathlib2 import Path  # python 2

from click.testing import CliRunner
from genbadge.main import genbadge as genbadge_cmd


TESTS_FOLDER = Path(__file__).parent.absolute()


class CmdReference:
    """
    Container class used to store our test reference for all commands
    """
    def __init__(self, name, default_infile, default_outfile, example_input_file, example_output_msg,
                 example_output_msg_long, help_msg):
        self.name = name
        self.default_infile = default_infile
        self.default_outfile = default_outfile
        self.example_input_file = example_input_file
        self.example_output_msg = example_output_msg
        self.example_output_msg_long = example_output_msg_long
        self.help_msg = help_msg

    def __str__(self):
        return self.name


TEST_CMD = CmdReference(
        name="tests",
        default_infile="reports/junit/junit.xml",
        default_outfile = "tests-badge.svg",
        example_input_file=(TESTS_FOLDER / "reports" / "junit" / "junit.xml").as_posix(),
        example_output_msg="SUCCESS - Tests badge created: %r\n",
        example_output_msg_long="""
Test statistics parsed successfully from %r
 - Nb tests: Total (6) = Success (2) + Skipped (1) + Failed (2) + Errors (1)
 - Success percentage: 40.00%% (2 / 5) (Skipped tests are excluded)

SUCCESS - Tests badge created: %r
""",
        help_msg="""Usage: genbadge tests [OPTIONS]

  This command generates a badge for the test results, from an XML file in the
  junit format. Such a file can be for example generated from python pytest
  using the --junitxml flag, or from java junit.

  By default the input file is the relative `./reports/junit/junit.xml` and the
  output file is `./tests-badge.svg`. You can change these settings with the
  `-i/--input_file` and `-o/--output-file` options.

  By default the badge will have the name "tests" as the left-hand side text.
  You can change these settings with the `-n/--name` option. The left-hand side
  text can be left blank with `-n ""` or have the left-hand side of the badge
  completely removed by passing `--noname`.

  You can use the verbose flag `-v/--verbose` to display information on the
  input file contents, for verification.

  The resulting badge will by default look like this: [tests | 6/12] where 6 is
  the number of tests that have run successfully, and 12 is the total number of
  tests minus the number of skipped tests. When all tests pass with success, the
  badge simply shows the number of tests [tests | 12]. You can change the
  appearance of the badge with the --format option (not implemented, todo).

  The success percentage is defined as 6/12 = 50.0%. You can use the
  `-t/--threshold` flag to setup a minimum success percentage required. If the
  success percentage is below the threshold, an error will be raised and the
  badge will not be generated.

Options:
  -i, --input-file FILENAME       An alternate test results XML file to read.
                                  '-' is supported and means <stdin>.
  -o, --output-file FILENAME      An alternate SVG badge file to write to. '-'
                                  is supported and means <stdout>. Note that in
                                  this case no other message will be printed to
                                  <stdout>. In particular the verbose flag will
                                  have no effect.
  -n, --name TEXT                 An alternate SVG badge text name to display on
                                  the left-hand side of the badge.
  -t, --threshold FLOAT           An optional success percentage threshold to
                                  use. The command will fail with exit code 1 if
                                  theactual success percentage is strictly less
                                  than the provided value.
  --withname / --noname           Indicates if a badge should be generated with
                                  or without the left-hand side of the badge.
  -w, --webshields / -l, --local  Indicates if badges should be generated using
                                  the shields.io HTTP API (default) or the local
                                  SVG file template included.
  -v, --verbose                   Use this flag to print details to stdout
                                  during the badge generation process. Note that
                                  this flag has no effect when '-' is used as
                                  output, since the badge is written to
                                  <stdout>. It also has no effect when the
                                  silent flag `-s` is used.
  -s, --silent                    When this flag is active nothing will be
                                  written to stdout. Note that this flag has no
                                  effect when '-' is used as the output file.
  --help                          Show this message and exit.
"""
)
COV_CMD = CmdReference(
        name="coverage",
        default_infile="reports/coverage/coverage.xml",
        default_outfile = "coverage-badge.svg",
        example_input_file=(TESTS_FOLDER / "reports" / "coverage" / "coverage.xml").as_posix(),
        example_output_msg="SUCCESS - Coverage badge created: %r\n",
        example_output_msg_long="""
Coverage results parsed successfully from %r
 - Branch coverage: 5.56%% (1/18)
 - Line coverage: 17.81%% (13/73)
 - Total coverage: 15.38%% ((1+13)/(18+73))

SUCCESS - Coverage badge created: %r
""",
        help_msg="""Usage: genbadge coverage [OPTIONS]

  This command generates a badge for the coverage results, from an XML file in
  the 'coverage' format. Such a file can be for example generated using the
  python `coverage` tool, or java `cobertura`.

  By default the input file is the relative `./reports/coverage/coverage.xml`
  and the output file is `./coverage-badge.svg`. You can change these settings
  with the `-i/--input_file` and `-o/--output-file` options.

  By default the badge will have the name "coverage" as the left-hand side text.
  You can change these settings with the `-n/--name` option. The left-hand side
  text can be left blank with `-n ""` or have the left-hand side of the badge
  completely removed by passing `--noname`.

  You can use the verbose flag `-v/--verbose` to display information on the
  input file contents, for verification.

  The resulting badge will by default look like this: [coverage | 98.1%] where
  98.1 is the total coverage, obtained from the branch and line coverages using
  the formula

      (nb_lines_covered + nb_branches_covered) / (nb_lines / nb_branches)

  and multiplying this by 100.

Options:
  -i, --input-file FILENAME       An alternate coverage results XML file to
                                  read. '-' is supported and means <stdin>.
  -o, --output-file FILENAME      An alternate SVG badge file to write to. '-'
                                  is supported and means <stdout>. Note that in
                                  this case no other message will be printed to
                                  <stdout>. In particular the verbose flag will
                                  have no effect.
  -n, --name TEXT                 An alternate SVG badge text name to display on
                                  the left-hand side of the badge.
  --withname / --noname           Indicates if a badge should be generated with
                                  or without the left-hand side of the badge.
  -w, --webshields / -l, --local  Indicates if badges should be generated using
                                  the shields.io HTTP API (default) or the local
                                  SVG file template included.
  -v, --verbose                   Use this flag to print details to stdout
                                  during the badge generation process. Note that
                                  this flag has no effect when '-' is used as
                                  output, since the badge is written to
                                  <stdout>. It also has no effect when the
                                  silent flag `-s` is used.
  -s, --silent                    When this flag is active nothing will be
                                  written to stdout. Note that this flag has no
                                  effect when '-' is used as the output file.
  --help                          Show this message and exit.
"""
)
FLAKE8_CMD = CmdReference(
        name="flake8",
        default_infile="reports/flake8/flake8stats.txt",
        default_outfile = "flake8-badge.svg",
        example_input_file=(TESTS_FOLDER / "reports" / "flake8" / "flake8stats.txt").as_posix(),
        example_output_msg="SUCCESS - Flake8 badge created: %r\n",
        example_output_msg_long="""
Flake8 statistics parsed successfully from %r
 - Total (20) = Critical (6) + Warning (9) + Info (5)

SUCCESS - Flake8 badge created: %r
""",
        help_msg="""Usage: genbadge flake8 [OPTIONS]

  This command generates a badge for the flake8 results, from a flake8stats.txt
  file. Such a file can be generated from python `flake8` using the --statistics
  flag.

  By default the input file is the relative `./reports/flake8/flake8stats.txt`
  and the output file is `./flake8-badge.svg`. You can change these settings
  with the `-i/--input_file` and `-o/--output-file` options.

  By default the badge will have the name "flake8" as the left-hand side text.
  You can change these settings with the `-n/--name` option. The left-hand side
  text can be left blank with `-n ""` or have the left-hand side of the badge
  completely removed by passing `--noname`.

  You can use the verbose flag `-v/--verbose` to display information on the
  input file contents, for verification.

  The resulting badge will by default look like this: [flake8 | 6 C, 0 W, 5 I]
  where 6, 0, 5 denote the number of critical issues, warnings, and information
  messages respectively. These severity levels are determined by the flake8-html
  plugin so as to match the colors in the HTML report. You can change the
  appearance of the badge with the --format option (not implemented, todo).

Options:
  -i, --input-file FILENAME       An alternate flake8 results TXT file to read.
                                  '-' is supported and means <stdin>.
  -o, --output-file FILENAME      An alternate SVG badge file to write to. '-'
                                  is supported and means <stdout>. Note that in
                                  this case no other message will be printed to
                                  <stdout>. In particular the verbose flag will
                                  have no effect.
  -n, --name TEXT                 An alternate SVG badge text name to display on
                                  the left-hand side of the badge.
  --withname / --noname           Indicates if a badge should be generated with
                                  or without the left-hand side of the badge.
  -w, --webshields / -l, --local  Indicates if badges should be generated using
                                  the shields.io HTTP API (default) or the local
                                  SVG file template included.
  -v, --verbose                   Use this flag to print details to stdout
                                  during the badge generation process. Note that
                                  this flag has no effect when '-' is used as
                                  output, since the badge is written to
                                  <stdout>. It also has no effect when the
                                  silent flag `-s` is used.
  -s, --silent                    When this flag is active nothing will be
                                  written to stdout. Note that this flag has no
                                  effect when '-' is used as the output file.
  --help                          Show this message and exit.
"""
)

ALL_COMMANDS = [TEST_CMD, COV_CMD, FLAKE8_CMD]
ALL_COMMAND_NAMES = [c.name for c in ALL_COMMANDS]


def test_help():
    """Test that `genbadge` provides the right help"""

    runner = CliRunner()
    result = runner.invoke(genbadge_cmd, [], catch_exceptions=False)

    assert result.exit_code == 0
    print(result.output)
    expected = """
Usage: genbadge [OPTIONS] COMMAND [ARGS]...

  Commandline utility to generate badges. To get help on each command use:

      genbadge <cmd> --help

Options:
  --help  Show this message and exit.

Commands:
  coverage  Generate a badge for the coverage results (e.g. from a
            coverage.xml).%s
  flake8    Generate a badge for the flake8 results (e.g. from a flake8stats.txt
            file).%s
  tests     Generate a badge for the test results (e.g. from a junit.xml).
"""
    if LooseVersion(click.__version__) < "8.":
        expected = expected % ("\n", "\n")
    else:
        expected = expected % ("", "")
    assert "\n" + result.output == expected


@pytest.mark.parametrize("cmd", ALL_COMMANDS, ids=str)
def test_help_cmd(cmd):
    """Test that the command-specific help message is correct"""

    result = _invoke_genbadge([cmd.name, "--help"])
    assert result.exit_code == 0

    if LooseVersion(click.__version__) >= "8.":
        assert result.output == cmd.help_msg
    else:
        # the line wrapping seems to have changed, do not check for this old version.
        pass


@pytest.mark.parametrize("cmd", ALL_COMMANDS, ids=str)
def test_file_not_found(monkeypatch, tmpdir, cmd):
    """Test that the error message is nice when the input file is not found"""

    currentfolder = Path(str(tmpdir))
    monkeypatch.chdir(str(currentfolder))

    # a) default input file: the error is raised by us as a click.exceptions.FileError (exit code 1)
    result = _invoke_genbadge([cmd.name])
    assert result.exit_code == 1
    expected = """
Error: Could not open file %r: File not found
"""
    # support various versions of click
    if LooseVersion(click.__version__) < "8.":
        expected = expected.replace("%r", "%s")
    assert "\n" + result.output == expected % cmd.default_infile

    # b) different non-existent input file: the error is raised by click from click.File as a BadParameterError (code 2)
    unknown_file = "unknown.file"
    result = _invoke_genbadge([cmd.name, "-i", unknown_file])
    assert result.exit_code == 2
    expected = """
Usage: genbadge {name} [OPTIONS]
Try 'genbadge {name} --help' for help.

Error: Invalid value for '-i' / '--input-file': %s: No such file or directory
""".format(name=cmd.name)
    # support various versions of click
    if LooseVersion(click.__version__) < "8.":
        expected = expected % "Could not open file: %s"
    else:
        expected = expected % "%r"
    assert "\n" + result.output == expected % unknown_file


@pytest.mark.parametrize("outstream", [False, True], ids="outstream={}".format)
@pytest.mark.parametrize("silent", [False, True], ids="silent={}".format)
@pytest.mark.parametrize("verbose", [False, True], ids="verbose={}".format)
@pytest.mark.parametrize("variant", ["default", "custom", "custom_shortargs", "custom_absolute"])
@pytest.mark.parametrize("cmd", ALL_COMMANDS, ids=str)
def test_any_command(monkeypatch, cmd, tmpdir, variant, outstream, silent, verbose):
    """Test that `genbadge <cmd>` works consistently concerning the ios and output messages"""

    # from pytest path to pathlib path
    currentfolder = Path(str(tmpdir))  # tests_folder

    # change the working directory to tmpdir
    monkeypatch.chdir(str(currentfolder))

    # create the various arguments. Use local template for faster exec
    args = [cmd.name, "-l"]
    if verbose:
        args.append("--verbose")
    if silent:
        args.append("--silent")
    if variant == "default":
        if outstream:
            pytest.skip("this test does not make sense")
        infile = currentfolder / cmd.default_infile
        outfile = currentfolder / cmd.default_outfile
        infile_path_for_msg = str(infile.absolute().as_posix())
    elif variant in ("custom", "custom_shortargs", "custom_absolute"):
        shortargs = "shortargs" in variant
        absolute = "absolute" in variant
        infile_path_for_msg = "foo/foo.xml"
        infile = currentfolder / infile_path_for_msg
        if absolute:
            infile_path_for_msg = infile.absolute().as_posix()
        outfile = currentfolder / "bar" / "bar-badge.svg"
        args += ["-i" if shortargs else "--input-file", infile_path_for_msg]
        args += ["-o" if shortargs else "--output-file", "-" if outstream else "bar/bar-badge.svg"]
    else:
        raise ValueError(variant)
    outfile_path_for_msg = str(outfile.absolute().as_posix())

    # copy the file from source
    infile.parent.mkdir(parents=True, exist_ok=True)
    copy(str(cmd.example_input_file), str(infile))

    # execute "genbadge tests" with the appropriate arguments
    print("Executing command in tmp folder %s" % (currentfolder,))
    result = _invoke_genbadge(args)
    assert result.exit_code == 0

    # verify the output message
    if not outstream:
        if silent:
            assert result.output == ""
        elif verbose:
            assert "\n" + result.output == cmd.example_output_msg_long % (infile_path_for_msg, outfile_path_for_msg)
        else:
            assert result.output == cmd.example_output_msg % outfile_path_for_msg
        assert outfile.exists()
    else:
        assert result.output.startswith('<svg xmlns="http://www.w3.org/2000/svg" '
                                        'xmlns:xlink="http://www.w3.org/1999/xlink" width=')


@pytest.mark.parametrize("threshold", [-1, 0, 40, 40.1, 100, 101])
@pytest.mark.parametrize("shortarg", [False, True])
def test_threshold(threshold, shortarg, tmpdir):

    # from pytest path to pathlib path
    destfolder = Path(str(tmpdir))
    badge_path = destfolder / "tests-badge.svg"

    # define cli args (explicit input so that we do not fall into the python 2 issue with CliRunner IOError
    args = ["tests", "-v", "-i", str(TEST_CMD.example_input_file), "-o", "%s" % badge_path]
    args += ["-t" if shortarg else "--threshold", str(threshold)]

    # execute "genbadge tests" with the appropriate arguments
    result = _invoke_genbadge(args)

    if 40.0 < threshold:
        assert result.exit_code == 1

        # verify the output message
        assert "\n" + result.output == """
Test statistics parsed successfully from '{}'
 - Nb tests: Total (6) = Success (2) + Skipped (1) + Failed (2) + Errors (1)
 - Success percentage: 40.00% (2 / 5) (Skipped tests are excluded)

Error: Success percentage 40.0% is strictly lower than required threshold {}%
""".format(str(TEST_CMD.example_input_file), float(threshold))

        assert not badge_path.exists()

    else:
        assert result.exit_code == 0

        # verify the output message
        assert "\n" + result.output == TEST_CMD.example_output_msg_long % (str(TEST_CMD.example_input_file), str(badge_path.as_posix()))

        assert badge_path.exists()


@pytest.mark.parametrize("use_shields,shortarg",
                         [(None, None), (False, False), (False, True), (True, False), (True, True)])
def test_local_remote(use_shields, shortarg, tmpdir):

    if use_shields is False and sys.version_info < (3,) and platform.system != "Windows":
        pytest.skip("On Linux the embedded ttf font file is needed, and because of the path change pkg_resources does"
                    "not manage to find the file on python 2")

    # from pytest path to pathlib path
    destfolder = Path(str(tmpdir))
    badge_path = destfolder / "tests-badge.svg"

    # define cli args (explicit input so that we do not fall into the python 2 issue with CliRunner IOError
    args = ["tests", "-v", "-i", str(TEST_CMD.example_input_file), "-o", "%s" % badge_path]
    if use_shields is False:
        args.append("-l" if shortarg else "--local")
    if use_shields is True:
        args.append("-w" if shortarg else "--webshields")

    # execute "genbadge tests" with the appropriate arguments
    result = _invoke_genbadge(args)
    assert result.exit_code == 0

    # verify the output message
    assert "\n" + result.output == TEST_CMD.example_output_msg_long % (str(TEST_CMD.example_input_file), str(badge_path.as_posix()))

    assert badge_path.exists()


def _invoke_genbadge(args):
    runner = CliRunner()
    print("\n> genbadge %s" % (" ".join(args),))
    result = runner.invoke(genbadge_cmd, args, catch_exceptions=False)
    print(result.output)
    return result
